# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from .basic_relations import *
from .detector import Detector
from .injections import *
from .multi_network import MultiNetwork
from .network import Network
from .psd import noise_curves_built_in
from .waveform import Waveform
from .utils import det_specs_built_in

# available detector locations
available_locs = list(det_specs_built_in.keys())

# available detector technologies
available_tecs = list(noise_curves_built_in.keys())
